#ifndef TRUTHMATCHSVC_CPP
#define TRUTHMATCHSVC_CPP

#include "TruthMatchingSvc.hpp"

#include "ConstDef.hpp"
#include "CutDefRX.hpp"

#include "HelperSvc.hpp"
#include "MessageSvc.hpp"

#include "CutHolder.hpp"
#include "EventType.hpp"

#include "TCut.h"

#include "vec_extends.h"
#include <fmt_ostream.h>

TCut MatchID(TString _particleLabel, int ID) {
    TString _matching = TString(fmt::format("TMath::Abs({0}_{1}) == {2}", _particleLabel.Data(), TRUEID, ID));
    return TCut(_matching);
}

TCut MatchIDWithGhost(TString _particleLabel, int ID) {
    TString _matching = TString(fmt::format("TMath::Abs({0}_{1}) == {2}", _particleLabel.Data(), TRUEID, ID));
    TString _isFake   = TString(fmt::format("TMath::Abs({0}_{1}) == 0", _particleLabel.Data(), TRUEID));
    return TCut(_matching) || TCut(_isFake);
}

TCut MatchMotherID(TString _particleLabel, int ID) {
    TString _matching = TString(fmt::format("TMath::Abs({0}_{1}) == {2}", _particleLabel.Data(), MOTHERID, ID));
    return TCut(_matching);
}

TCut MatchGrandMotherID(TString _particleLabel, int ID) {
    TString _matching = TString(fmt::format("TMath::Abs({0}_{1}) == {2}", _particleLabel.Data(), GMOTHERID, ID));
    return TCut(_matching);
}

TCut MatchGrandGrandMotherID(TString _particleLabel, int ID) {
    TString _matching = TString(fmt::format("TMath::Abs({0}_{1}) == {2}", _particleLabel.Data(), GGMOTHERID, ID));
    return TCut(_matching);
}

TCut TruthMatching(const Prj & _prj, const Analysis & _ana, const Q2Bin & _q2bin, const TString & _sample, TString _option, bool _debug, bool _swap) {
    if (!CheckVectorContains(SettingDef::AllowedConf::Samples.at(_prj), _sample)) {
        MessageSvc::Error("TruthMatching", to_string(_prj), to_string(_ana), _sample, "not in allowed config", "EXIT_FAILURE");
        return TCut(NOCUT);
    }

    if (_debug) MessageSvc::Debug("TruthMatching", to_string(_prj), to_string(_ana), _sample, to_string(_swap));

    TCut _cut(NOCUT);

    // We label the di-lepton intermediate state as JPs, regardless of q2 and charmonium state
    TString _head, _hh, _h1, _h2, _ll;
    switch (_prj) {
        case Prj::RKst:
            _head = "B0";
            _hh   = "Kst";
            _h1   = "K";
            _h2   = "Pi";
            _ll = "JPs";
            break;
        case Prj::RK:
            _head = "Bp";
            _hh   = "";
            _h1   = "K";
            _h2   = "";
            _ll = "JPs";
            break;
        case Prj::RPhi:
            _head = "Bs";
            _hh   = "Phi";
            _h1   = "K1";
            _h2   = "K2";
            _ll = "JPs";
            break;
        default: MessageSvc::Error("TruthMatching", (TString) "Invalid project", to_string(_prj), "EXIT_FAILURE"); break;
    }

    TString _l1, _l2;
    int     _l1ID = 0, _l2ID = 0;
    switch (_ana) {
        case Analysis::MM:
            _l1   = "M1";
            _l2   = "M2";
            _l1ID = PDG::ID::M;
            _l2ID = PDG::ID::M;
            break;
        case Analysis::EE:
            _l1   = "E1";
            _l2   = "E2";
            _l1ID = PDG::ID::E;
            _l2ID = PDG::ID::E;
            break;
        case Analysis::ME:
            _l1   = "M1";
            _l2   = "E2";
            _l1ID = PDG::ID::M;
            _l2ID = PDG::ID::E;
            break;
        default: MessageSvc::Error("TruthMatching", (TString) "Invalid analysis", to_string(_ana), "EXIT_FAILURE"); break;
    }

    int _headID = 0;
    if (_sample.BeginsWith("Bd")) _headID = PDG::ID::Bd;
    if (_sample.BeginsWith("Bu")) _headID = PDG::ID::Bu;
    if (_sample.BeginsWith("Bs")) _headID = PDG::ID::Bs;
    if (_sample.BeginsWith("Lb")) _headID = PDG::ID::Lb;
    if (_headID == 0) MessageSvc::Error("TruthMatching", (TString) "Invalid HEAD", _sample, to_string(_headID), "EXIT_FAILURE");

    TCut _headMatched = MatchID(_head, _headID);

    TCut _finalStatesMatchedHH      = TCut(NOCUT);
    TCut _finalStatesMatchedHH_SWAP = TCut(NOCUT);
    if (_sample.Contains("2Pi")) { _finalStatesMatchedHH = MatchID(_h1, PDG::ID::Pi); }
    if (_sample.Contains("2K")) { _finalStatesMatchedHH = MatchID(_h1, PDG::ID::K); }
    if (_sample.Contains("2Kst") || _sample.Contains("2KPi")) {
        _finalStatesMatchedHH      = MatchID(_h1, PDG::ID::K) && MatchID(_h2, PDG::ID::Pi);
        _finalStatesMatchedHH_SWAP = MatchID(_h1, PDG::ID::Pi) && MatchID(_h2, PDG::ID::K);
        if (_prj == Prj::RK) {
            _finalStatesMatchedHH      = MatchID(_h1, PDG::ID::K);
            _finalStatesMatchedHH_SWAP = MatchID(_h1, PDG::ID::Pi);
        }
    }
    if (_sample.Contains("2Phi")) {
        _finalStatesMatchedHH = MatchID(_h1, PDG::ID::K) && MatchID(_h2, PDG::ID::K);
        if (_prj == Prj::RK) { _finalStatesMatchedHH = MatchID(_h1, PDG::ID::K); }
    }
    if (_sample.Contains("2pK")) {
        _finalStatesMatchedHH      = MatchID(_h1, PDG::ID::K) && MatchID(_h2, PDG::ID::P);
        _finalStatesMatchedHH_SWAP = MatchID(_h1, PDG::ID::P) && MatchID(_h2, PDG::ID::K);
        if (_prj == Prj::RK) {
            _finalStatesMatchedHH      = MatchID(_h1, PDG::ID::K);
            _finalStatesMatchedHH_SWAP = MatchID(_h1, PDG::ID::P);
        }
    }
    if (_sample.Contains("2X")) {
        if (_prj == Prj::RKst) {
            _finalStatesMatchedHH      = MatchID(_h1, PDG::ID::K) && MatchID(_h2, PDG::ID::Pi);
            _finalStatesMatchedHH_SWAP = (MatchID(_h1, PDG::ID::Pi) && MatchID(_h2, PDG::ID::K)) || (MatchID(_h1, PDG::ID::K) && MatchID(_h2, PDG::ID::K)) || (MatchID(_h1, PDG::ID::Pi) && MatchID(_h2, PDG::ID::Pi)) || (MatchID(_h1, PDG::ID::K) && MatchID(_h2, PDG::ID::P)) || (MatchID(_h1, PDG::ID::P) && MatchID(_h2, PDG::ID::K));
        }
        if (_prj == Prj::RK) {
            _finalStatesMatchedHH      = MatchID(_h1, PDG::ID::K);
            _finalStatesMatchedHH_SWAP = MatchID(_h1, PDG::ID::Pi) || MatchID(_h1, PDG::ID::P);
        }
    }
    if (_swap && (_finalStatesMatchedHH_SWAP != TCut(NOCUT))) _finalStatesMatchedHH = _finalStatesMatchedHH || _finalStatesMatchedHH_SWAP;

    TCut _finalStatesMatchedLL = MatchID(_l1, _l1ID) && MatchID(_l2, _l2ID);

    // TCut _finalStatesMatched2X = _finalStatesMatchedHH && _finalStatesMatchedLL;
    TCut _finalStatesMatched2X = TCut(NOCUT);
    if (_prj == Prj::RK) {
        _finalStatesMatched2X = !MatchID(_h1, _headID) && !MatchID(_ll, _headID) && !MatchID(_l1, _headID) && !MatchID(_l2, _headID);
    } else {
        _finalStatesMatched2X = !MatchID(_hh, _headID) && !MatchID(_h1, _headID) && !MatchID(_h2, _headID) && !MatchID(_ll, _headID) && !MatchID(_l1, _headID) && !MatchID(_l2, _headID);
    }
    
    if (_prj == Prj::RKst) {
        if (_sample == "Bd2KstMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("M1", PDG::ID::Bd) && MatchMotherID("M2", PDG::ID::Bd) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchGrandMotherID("Pi", PDG::ID::Bd); }
        if (_sample == "Bd2KstEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("E1", PDG::ID::Bd) && MatchMotherID("E2", PDG::ID::Bd) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchGrandMotherID("Pi", PDG::ID::Bd); }
        if (_sample == "Bd2KstJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchGrandMotherID("Pi", PDG::ID::Bd) && MatchGrandMotherID("M1", PDG::ID::Bd) && MatchGrandMotherID("M2", PDG::ID::Bd); }
        if (_sample == "Bd2KstSwapJPsMM") { _cut = _headMatched && _finalStatesMatchedHH_SWAP && _finalStatesMatchedLL && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("M1", PDG::ID::Bd) && MatchGrandMotherID("M2", PDG::ID::Bd); }
        if (_sample == "Bd2KstJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchGrandMotherID("Pi", PDG::ID::Bd) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }
        if (_sample == "Bd2KstSwapJPsEE") { _cut = _headMatched && _finalStatesMatchedHH_SWAP && _finalStatesMatchedLL && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }
        if (_sample == "Bd2KstPsiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchGrandMotherID("Pi", PDG::ID::Bd) && MatchGrandMotherID("M1", PDG::ID::Bd) && MatchGrandMotherID("M2", PDG::ID::Bd); }
        if (_sample == "Bd2KstSwapPsiMM") { _cut = _headMatched && _finalStatesMatchedHH_SWAP && _finalStatesMatchedLL && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("M1", PDG::ID::Bd) && MatchGrandMotherID("M2", PDG::ID::Bd); }
        if (_sample == "Bd2KstPsiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchGrandMotherID("Pi", PDG::ID::Bd) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }
        if (_sample == "Bd2KstSwapPsiEE") { _cut = _headMatched && _finalStatesMatchedHH_SWAP && _finalStatesMatchedLL && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }
        
        if (_sample == "Bd2KPiJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bd) && MatchMotherID("Pi", PDG::ID::Bd) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("M1", PDG::ID::Bd) && MatchGrandMotherID("M2", PDG::ID::Bd); }
        if (_sample == "Bd2KPiJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bd) && MatchMotherID("Pi", PDG::ID::Bd) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }

        if (_sample == "Bd2KstPsiPiPiJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchGrandMotherID("Pi", PDG::ID::Bd) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Psi) && MatchGrandMotherID("E2", PDG::ID::Psi); }
        
        if (_sample == "Bu2KMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchGrandMotherID("M1", PDG::ID::Bu) && MatchGrandMotherID("M2", PDG::ID::Bu); }
        if (_sample == "Bu2KEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchGrandMotherID("E1", PDG::ID::Bu) && MatchGrandMotherID("E2", PDG::ID::Bu); }
        if (_sample == "Bu2KJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("M1", PDG::ID::Bu) && MatchGrandMotherID("M2", PDG::ID::Bu); }
        if (_sample == "Bu2KJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bu) && MatchGrandMotherID("E2", PDG::ID::Bu); }
        if (_sample == "Bu2KPsiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("M1", PDG::ID::Bu) && MatchGrandMotherID("M2", PDG::ID::Bu); }
        if (_sample == "Bu2KPsiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("E1", PDG::ID::Bu) && MatchGrandMotherID("E2", PDG::ID::Bu); }

        if (_sample == "Bd2KPiPsiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bd) && MatchMotherID("Pi", PDG::ID::Bd) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("M1", PDG::ID::Bd) && MatchGrandMotherID("M2", PDG::ID::Bd); }
        if (_sample == "Bd2KPiPsiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bd) && MatchMotherID("Pi", PDG::ID::Bd) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }

        if (_sample == "Bs2KstJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }
        if (_sample == "Bs2KstJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }

        if (_sample == "Bs2KstPsiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }
        if (_sample == "Bs2KstPsiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }

        if (_sample == "Bs2KPiJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bs) && MatchMotherID("Pi", PDG::ID::Bs) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }
        if (_sample == "Bs2KPiJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bs) && MatchMotherID("Pi", PDG::ID::Bs) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }

        if (_sample == "Bs2PhiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Phi) && MatchMotherID("Pi", PDG::ID::Phi) && MatchMotherID("E1", PDG::ID::Bs) && MatchMotherID("E2", PDG::ID::Bs) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs); }
        if (_sample == "Bs2PhiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Phi) && MatchMotherID("Pi", PDG::ID::Phi) && MatchMotherID("M1", PDG::ID::Bs) && MatchMotherID("M2", PDG::ID::Bs) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs); }
        if (_sample == "Bs2PhiJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Phi) && MatchMotherID("Pi", PDG::ID::Phi) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Phi) && MatchMotherID("Pi", PDG::ID::Phi) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiPsiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Phi) && MatchMotherID("Pi", PDG::ID::Phi) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiPsiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Phi) && MatchMotherID("Pi", PDG::ID::Phi) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("Pi", PDG::ID::Bs) && MatchGrandMotherID("K", PDG::ID::Bs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }


        if (_sample == "Lb2pKMM") { _cut = _headMatched && (_finalStatesMatchedHH_SWAP || _finalStatesMatchedHH) && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Lb) && MatchMotherID("Pi", PDG::ID::Lb) && MatchMotherID("M1", PDG::ID::Lb) && MatchMotherID("M2", PDG::ID::Lb); }
        if (_sample == "Lb2pKEE") { _cut = _headMatched && (_finalStatesMatchedHH_SWAP || _finalStatesMatchedHH) && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Lb) && MatchMotherID("Pi", PDG::ID::Lb) && MatchMotherID("E1", PDG::ID::Lb) && MatchMotherID("E2", PDG::ID::Lb); }
        if (_sample == "Lb2pKJPsMM") { _cut = _headMatched && (_finalStatesMatchedHH_SWAP || _finalStatesMatchedHH) && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Lb) && MatchMotherID("Pi", PDG::ID::Lb) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("M1", PDG::ID::Lb) && MatchGrandMotherID("M2", PDG::ID::Lb); }
        if (_sample == "Lb2pKJPsEE") { _cut = _headMatched && (_finalStatesMatchedHH_SWAP || _finalStatesMatchedHH) && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Lb) && MatchMotherID("Pi", PDG::ID::Lb) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Lb) && MatchGrandMotherID("E2", PDG::ID::Lb); }
        if (_sample == "Lb2pKPsiMM") { _cut = _headMatched && (_finalStatesMatchedHH_SWAP || _finalStatesMatchedHH) && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Lb) && MatchMotherID("Pi", PDG::ID::Lb) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("M1", PDG::ID::Lb) && MatchGrandMotherID("M2", PDG::ID::Lb); }
        if (_sample == "Lb2pKPsiEE") { _cut = _headMatched && (_finalStatesMatchedHH_SWAP || _finalStatesMatchedHH) && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Lb) && MatchMotherID("Pi", PDG::ID::Lb) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("E1", PDG::ID::Lb) && MatchGrandMotherID("E2", PDG::ID::Lb); }

        if (_sample == "Bu2K1EE") {
            // see here : https://gitlab.cern.ch/lhcb-datapkg/Gen/DecFiles/blob/master/dkfiles/Bu_K1ee=DecProdCut.dec
            // B+ -> ... K1(1270)+ -> (K*0 -> K+pi-)pi+
            TCut _kstPi_match = MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) &&   // want to truth match also Mis-ID case? (not sure)
                                MatchGrandMotherID("K", PDG::ID::K_1_1270_c) && MatchGrandMotherID("Pi", PDG::ID::K_1_1270_c) && MatchGrandGrandMotherID("K", PDG::ID::Bu) && MatchGrandGrandMotherID("Pi", PDG::ID::Bu);
            // B+ -> ... K(1270)+ -> (rho -> pi+pi-)K+
            TCut _rhoK_match = MatchMotherID("Pi", PDG::ID::Rho0) && MatchMotherID("K", PDG::ID::K_1_1270_c) && MatchGrandMotherID("Pi", PDG::ID::K_1_1270_c) && MatchGrandMotherID("K", PDG::ID::Bu) && MatchGrandGrandMotherID("Pi", PDG::ID::Bu);

            TCut _omega_match = MatchMotherID("Pi", PDG::ID::Omega) && MatchMotherID("K", PDG::ID::K_1_1270_c) && MatchGrandMotherID("Pi", PDG::ID::K_1_1270_c) && MatchGrandMotherID("K", PDG::ID::Bu) && MatchGrandGrandMotherID("Pi", PDG::ID::Bu);
            TCut _finalCut    = (_omega_match || _rhoK_match || _kstPi_match) && _finalStatesMatchedLL;
            _cut              = _finalCut;
        }
        if (_sample == "Bu2K2EE") {
            TCut _kstPi_match = MatchMotherID("K", PDG::ID::Kst) && MatchMotherID("Pi", PDG::ID::Kst) && MatchGrandMotherID("K", PDG::ID::K_2_1430_c) && MatchGrandMotherID("Pi", PDG::ID::K_2_1430_c) && MatchGrandGrandMotherID("K", PDG::ID::Bu) && MatchGrandGrandMotherID("Pi", PDG::ID::Bu);

            TCut _rhoK_match = MatchMotherID("Pi", PDG::ID::Rho0) && MatchMotherID("K", PDG::ID::K_2_1430_c) && MatchGrandMotherID("Pi", PDG::ID::K_2_1430_c) && MatchGrandMotherID("K", PDG::ID::Bu) && MatchGrandGrandMotherID("Pi", PDG::ID::Bu);

            TCut _omega_match = MatchMotherID("Pi", PDG::ID::Omega) && MatchMotherID("K", PDG::ID::K_2_1430_c) && MatchGrandMotherID("Pi", PDG::ID::K_2_1430_c) && MatchGrandMotherID("K", PDG::ID::Bu) && MatchGrandGrandMotherID("Pi", PDG::ID::Bu);
            TCut _finalCut    = (_omega_match || _rhoK_match || _kstPi_match) && _finalStatesMatchedLL;
            _cut              = _finalCut;
        }
        if (_sample == "Bu2KPiPiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("Pi", PDG::ID::Bu) && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::Bu) && MatchMotherID("E2", PDG::ID::Bu); }
        if (_sample == "Bu2KPiPiJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("Pi", PDG::ID::Bu) && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bu) && MatchGrandMotherID("E2", PDG::ID::Bu); }

        if (_sample.Contains("2XJPs")) {
            _cut = _headMatched && _finalStatesMatched2X && !TRUTHMATCHING::bkgCatSignal;
            if (_q2bin == Q2Bin::JPsi){
                // Veto Bd2KstJPsLL and Bd2KPiLL
                if (_sample.BeginsWith("Bd")) _cut = _cut && !TruthMatching(_prj, _ana, _q2bin, "Bd2KstJPs" + to_string(_ana), _option, _debug, true) && !TruthMatching(_prj, _ana, _q2bin, "Bd2KPiJPs" + to_string(_ana), _option, _debug, true);
                // Veto Bs2KstJPs and Bs2PhiJPs (already included as dedicated component!)
                if (_sample.BeginsWith("Bs")) _cut = _cut && !TruthMatching(_prj, _ana, _q2bin, "Bs2KstJPs" + to_string(_ana), _option, _debug, true) && !TruthMatching(_prj, _ana, _q2bin, "Bs2PhiJPs" + to_string(_ana), _option, _debug, true);                
                TCut _lep = MatchGrandMotherID(_l1, _headID) && MatchGrandMotherID(_l2, _headID) && MatchMotherID(_l1, PDG::ID::JPs) && MatchMotherID(_l2, PDG::ID::JPs);
                if (_option.Contains("PRH")) _cut = _cut && _lep;
                if (_option.Contains("PRL")) _cut = _cut && !_lep;
            }
            else if (_q2bin == Q2Bin::Psi){
                if (_sample.BeginsWith("Bd")) {
            		//Veto Bd2KstPsiLL and Bd2KPiPsi
                    TCut _notPsiMode  = !TruthMatching(_prj, _ana, _q2bin, "Bd2KstPsi" + to_string(_ana), _option, _debug, true) && !TruthMatching(_prj, _ana, _q2bin, "Bd2KPiPsi" + to_string(_ana), _option, _debug, true);
        		    //Veto leakage
                    TCut _notJPsiMode = !TruthMatching(_prj, _ana, _q2bin, "Bd2KstJPs" + to_string(_ana), _option, _debug, true) && !TruthMatching(_prj, _ana, _q2bin, "Bd2KPiJPs" + to_string(_ana), _option, _debug, true);
                    _cut = _cut && _notPsiMode && _notJPsiMode;
                }
                //Veto Bs2KstPsi and Bs2PhiPsi (already included as dedicated component!)
                if (_sample.BeginsWith("Bs")) _cut = _cut && !TruthMatching(_prj, _ana, _q2bin, "Bs2KstPsi" + to_string(_ana), _option, _debug, true) && !TruthMatching(_prj, _ana, _q2bin, "Bs2PhiPsi" + to_string(_ana), _option, _debug, true);;
                // TODO ( not ready to veto this until we have new MC for Psi-> PiPIJPs Kst)
                TCut _psiToJPsi = MatchGrandMotherID(_l1, PDG::ID::Psi) && MatchGrandMotherID(_l2, PDG::ID::Psi) && MatchMotherID(_l1, PDG::ID::JPs) && MatchMotherID(_l2, PDG::ID::JPs);
                if (_ana == Analysis::EE) _cut = _cut && !_psiToJPsi;		
                /*
                        if (_option.Contains("PRH")) _cut = _cut && !_psiToJPsi;
                        if (_option.Contains("PRL")) _cut = _cut && _psiToJPsi;
                */
            }
        }
    }

    if (_prj == Prj::RK) {
        if (_sample == "Bu2KMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("M1", PDG::ID::Bu) && MatchMotherID("M2", PDG::ID::Bu); }
        if (_sample == "Bu2KEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::Bu) && MatchMotherID("E2", PDG::ID::Bu); }
        if (_sample == "Bu2KJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("M1", PDG::ID::Bu) && MatchGrandMotherID("M2", PDG::ID::Bu); }
        if (_sample == "Bu2KJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bu) && MatchGrandMotherID("E2", PDG::ID::Bu); }
        if (_sample == "Bu2KPsiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("M1", PDG::ID::Bu) && MatchGrandMotherID("M2", PDG::ID::Bu); }
        if (_sample == "Bu2KPsiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("E1", PDG::ID::Bu) && MatchGrandMotherID("E2", PDG::ID::Bu); }
        if (_sample == "Bu2KPsiPiPiJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Psi) && MatchGrandMotherID("E2", PDG::ID::Psi); }	

        if (_sample == "Bu2PiJPsMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchGrandMotherID("M1", PDG::ID::Bu) && MatchGrandMotherID("M2", PDG::ID::Bu); }
        if (_sample == "Bu2PiJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bu) && MatchGrandMotherID("E2", PDG::ID::Bu); }

        if (_sample == "Bu2KstEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst_c) && MatchGrandMotherID("K", PDG::ID::Bu) && MatchMotherID("E1", PDG::ID::Bu) && MatchMotherID("E2", PDG::ID::Bu); }

        if (_sample == "Bd2KstEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchMotherID("E1", PDG::ID::Bd) && MatchMotherID("E2", PDG::ID::Bd); }
        if (_sample == "Bd2KstJPsEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }
        if (_sample == "Bd2KstPsiEE") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchGrandMotherID("E1", PDG::ID::Bd) && MatchGrandMotherID("E2", PDG::ID::Bd); }
        if (_sample == "Bd2KstPsiMM") { _cut = _headMatched && _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("K", PDG::ID::Kst) && MatchGrandMotherID("K", PDG::ID::Bd) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchGrandMotherID("M1", PDG::ID::Bd) && MatchGrandMotherID("M2", PDG::ID::Bd); }	
        if (_sample.Contains("2XJPs")){
            _cut = _headMatched && _finalStatesMatched2X && !TRUTHMATCHING::bkgCatSignal;
            if (_q2bin == Q2Bin::JPsi){
                if (_sample.BeginsWith("Bu")) _cut = _cut && !TruthMatching(_prj, _ana, _q2bin, "Bu2KJPs" + to_string(_ana), _option, _debug, true) && !TruthMatching(_prj, _ana, _q2bin, "Bu2PiJPs" + to_string(_ana), _option, _debug, true);                                
                TCut _lep = MatchGrandMotherID(_l1, _headID) && MatchGrandMotherID(_l2, _headID) && MatchMotherID(_l1, PDG::ID::JPs) && MatchMotherID(_l2, PDG::ID::JPs);
                if (_option.Contains("PRH")) _cut = _cut && _lep;
                if (_option.Contains("PRL")) _cut = _cut && !_lep;
            }
            else if (_q2bin == Q2Bin::Psi){
                auto matchLMothers= [&]( const int & pdgID,   bool posLogic){ 
                    TCut _llmatch = MatchMotherID(_l1, pdgID) && MatchMotherID(_l2, pdgID);
                    return posLogic? _llmatch : !_llmatch; 
                };
                auto matchLGMothers= [&]( const int & pdgID,  bool posLogic){ 
                    TCut _llmatch = MatchGrandMotherID(_l1, pdgID) && MatchGrandMotherID(_l2, pdgID);
                    return posLogic? _llmatch : !_llmatch; 
                };
                auto matchLGGMothers= [&]( const int & pdgID, bool posLogic){ 
                    TCut _llmatch = MatchGrandGrandMotherID(_l1, pdgID) && MatchGrandGrandMotherID(_l2, pdgID);
                    return posLogic? _llmatch : !_llmatch; 
                };
                //JPsi -> LL 
                TCut _LL_from_JPs           = matchLMothers(PDG::ID::JPs, true);
                //Psi2S -> LL
                TCut _LL_from_Psi           = matchLMothers(PDG::ID::Psi, true);
                //B-> (Psi2S -> LL ) X 
                TCut _LL_from_Xc            = matchLMothers(PDG::ID::JPs, false) && matchLMothers( PDG::ID::Psi, false);
                //Psi->LL and Psi from B directly
                TCut _LL_from_Psi_PsifromB  = _LL_from_Psi && matchLGMothers(_headID,true);
                //JPs->LL and JPs from resonance which is not a Psi
                TCut _JPs_from_Psi          = _LL_from_JPs && matchLGMothers(PDG::ID::Psi, true);
                //JPsi->LL and JPsi from a Psi in cascade
                TCut _JPs_from_PsiCascade   = _LL_from_JPs && matchLGGMothers(PDG::ID::Psi, true);
                //JPsi->LL and JPsi from not a Psi or a Psi in cascade
                TCut _JPs_from_Other        = _LL_from_JPs && matchLGMothers(PDG::ID::Psi, false) && matchLGGMothers(PDG::ID::Psi, false);
                //Psi->LL and Psi from other
                TCut _Psi_from_Other        = _LL_from_Psi && ( matchLGMothers(_headID, false) || matchLGGMothers(_headID, false));
                //B-> H X RES->( Y LL)
                TCut _H_fromB               =  MatchMotherID(_h1, _headID);
                //B-> STR(->H X) RES->( Y LL)
                TCut _H_fromResonance       = !_H_fromB && ( MatchGrandMotherID(_h1,_headID ) || MatchGrandGrandMotherID(_h1,_headID));

                if (_sample.BeginsWith("Bu")){
                    //B+ -> K+(Y) (Z-> Psi/JPs->EE)
                    TCut _notSignalMode =  !TruthMatching(_prj, _ana, _q2bin, "Bu2KPsi" + to_string(_ana), _option, _debug, true); //remove         Bu2KPsi on RK
                    TCut _notJPsiMode   =  !TruthMatching(_prj, _ana, _q2bin, "Bu2KJPs" + to_string(_ana), _option, _debug, true); //remove leakage Bu2KJPs on RK
                    _cut = _cut && _notSignalMode; //also for Muon mode
                    TCut _isPartRecoH   = _LL_from_Psi_PsifromB && !_H_fromB;            
                    if (_ana == Analysis::EE){                        
                        _cut = _cut && _notSignalMode && _notJPsiMode; 
                        //We model the Psi-> XJPs with intermediate resonances with PiPi JPs
                        TCut _backgroundConsidered =  _JPs_from_Psi || _JPs_from_PsiCascade || _JPs_from_Other || _Psi_from_Other || _LL_from_Xc;
                        TCut _backgroundVetoed     = _H_fromB  && ( _JPs_from_Psi || _JPs_from_PsiCascade);
                        _cut = _cut && _backgroundConsidered && !_backgroundVetoed;
                    }
                }
                if( _sample.BeginsWith("Bd" && _ana == Analysis::EE )){
                    //B0 -> K+(Y-) (Z->Psi/Jps->EE)  , must veto PartReco only here i think 
                    TCut _kPartReco     = MatchMotherID(_h1, PDG::ID::Kst); //we wil keep everything which is has not a Psi->LL direct decay. Part-Reco from 1 missing particle will not show up i think .                     
                    TCut _backgroundConsidered =  (_JPs_from_Psi || _JPs_from_PsiCascade || _JPs_from_Other || _Psi_from_Other || _LL_from_Xc );
                    _cut = _cut && !_kPartReco && (_backgroundConsidered);
                }

                //remove Bu2Psi(->pipiJ/Psi)K since it's modelled!
                // TCut _psiToJPsi = MatchGrandMotherID(_l1, PDG::ID::Psi) && MatchGrandMotherID(_l2, PDG::ID::Psi) && MatchMotherID(_l1, PDG::ID::JPs) && MatchMotherID(_l2, PDG::ID::JPs); 
                /*
                if (_option.Contains("PRH")) _cut = _cut && !_psiToJPsi;
                if (_option.Contains("PRL")) _cut = _cut && _psiToJPsi;
                */
            }
        }
    }

    if (_prj == Prj::RPhi) {
        if (_sample == "Bs2PhiMM") { _cut = _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("M1", PDG::ID::Bs) && MatchMotherID("M2", PDG::ID::Bs) && MatchMotherID("K1", PDG::ID::Phi) && MatchMotherID("K2", PDG::ID::Phi) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiEE") { _cut = _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("E1", PDG::ID::Bs) && MatchMotherID("E2", PDG::ID::Bs) && MatchMotherID("K1", PDG::ID::Phi) && MatchMotherID("K2", PDG::ID::Phi) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiJPsMM") { _cut = _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchMotherID("K1", PDG::ID::Phi) && MatchMotherID("K2", PDG::ID::Phi) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiJPsEE") { _cut = _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchMotherID("K1", PDG::ID::Phi) && MatchMotherID("K2", PDG::ID::Phi) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiPsiMM") { _cut = _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchMotherID("K1", PDG::ID::Phi) && MatchMotherID("K2", PDG::ID::Phi) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }
        if (_sample == "Bs2PhiPsiEE") { _cut = _finalStatesMatchedHH && _finalStatesMatchedLL && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchMotherID("K1", PDG::ID::Phi) && MatchMotherID("K2", PDG::ID::Phi) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }

        if (_sample == "Bd2KstMM") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::Pi)) || (MatchID("K1", PDG::ID::Pi) && MatchID("K2", PDG::ID::K))) && MatchMotherID("M1", PDG::ID::Bs) && MatchMotherID("M2", PDG::ID::Bs) && MatchMotherID("K1", PDG::ID::Kst) && MatchMotherID("K2", PDG::ID::Kst) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs); }
        if (_sample == "Bd2KstEE") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::Pi)) || (MatchID("K1", PDG::ID::Pi) && MatchID("K2", PDG::ID::K))) && MatchMotherID("E1", PDG::ID::Bs) && MatchMotherID("E2", PDG::ID::Bs) && MatchMotherID("K1", PDG::ID::Kst) && MatchMotherID("K2", PDG::ID::Kst) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs); }
        if (_sample == "Bd2KstJPsMM") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::Pi)) || (MatchID("K1", PDG::ID::Pi) && MatchID("K2", PDG::ID::K))) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchMotherID("K1", PDG::ID::Kst) && MatchMotherID("K2", PDG::ID::Kst) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs) && MatchGrandMotherID("M1", PDG::ID::Bs) && MatchGrandMotherID("M2", PDG::ID::Bs); }
        if (_sample == "Bd2KstJPsEE") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::Pi)) || (MatchID("K1", PDG::ID::Pi) && MatchID("K2", PDG::ID::K))) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchMotherID("K1", PDG::ID::Kst) && MatchMotherID("K2", PDG::ID::Kst) && MatchGrandMotherID("K1", PDG::ID::Bs) && MatchGrandMotherID("K2", PDG::ID::Bs) && MatchGrandMotherID("E1", PDG::ID::Bs) && MatchGrandMotherID("E2", PDG::ID::Bs); }

        if (_sample == "Lb2pKMM") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::P)) || (MatchID("K1", PDG::ID::P) && MatchID("K2", PDG::ID::K))) && MatchMotherID("M1", PDG::ID::Lb) && MatchMotherID("M2", PDG::ID::Lb) && MatchMotherID("K1", PDG::ID::Lb) && MatchMotherID("K2", PDG::ID::Lb); }
        if (_sample == "Lb2pKEE") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::P)) || (MatchID("K1", PDG::ID::P) && MatchID("K2", PDG::ID::K))) && MatchMotherID("E1", PDG::ID::Lb) && MatchMotherID("E2", PDG::ID::Lb) && MatchMotherID("K1", PDG::ID::Lb) && MatchMotherID("K2", PDG::ID::Lb); }
        if (_sample == "Lb2pKJPsMM") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::P)) || (MatchID("K1", PDG::ID::P) && MatchID("K2", PDG::ID::K))) && MatchMotherID("M1", PDG::ID::JPs) && MatchMotherID("M2", PDG::ID::JPs) && MatchMotherID("K1", PDG::ID::Lb) && MatchMotherID("K2", PDG::ID::Lb) && MatchGrandMotherID("M1", PDG::ID::Lb) && MatchGrandMotherID("M2", PDG::ID::Lb); }
        if (_sample == "Lb2pKJPsEE") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::P)) || (MatchID("K1", PDG::ID::P) && MatchID("K2", PDG::ID::K))) && MatchMotherID("E1", PDG::ID::JPs) && MatchMotherID("E2", PDG::ID::JPs) && MatchMotherID("K1", PDG::ID::Lb) && MatchMotherID("K2", PDG::ID::Lb) && MatchGrandMotherID("E1", PDG::ID::Lb) && MatchGrandMotherID("E2", PDG::ID::Lb); }
        if (_sample == "Lb2pKPsiMM") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::P)) || (MatchID("K1", PDG::ID::P) && MatchID("K2", PDG::ID::K))) && MatchMotherID("M1", PDG::ID::Psi) && MatchMotherID("M2", PDG::ID::Psi) && MatchMotherID("K1", PDG::ID::Lb) && MatchMotherID("K2", PDG::ID::Lb) && MatchGrandMotherID("M1", PDG::ID::Lb) && MatchGrandMotherID("M2", PDG::ID::Lb); }
        if (_sample == "Lb2pKPsiEE") { _cut = _finalStatesMatchedLL && ((MatchID("K1", PDG::ID::K) && MatchID("K2", PDG::ID::P)) || (MatchID("K1", PDG::ID::P) && MatchID("K2", PDG::ID::K))) && MatchMotherID("E1", PDG::ID::Psi) && MatchMotherID("E2", PDG::ID::Psi) && MatchMotherID("K1", PDG::ID::Lb) && MatchMotherID("K2", PDG::ID::Lb) && MatchGrandMotherID("E1", PDG::ID::Lb) && MatchGrandMotherID("E2", PDG::ID::Lb); }

        if (_sample.Contains("2XJPs")) {
            _cut = _headMatched && _finalStatesMatched2X && !TRUTHMATCHING::bkgCatSignal;

            TCut _lep = MatchGrandMotherID(_l1, _headID) && MatchGrandMotherID(_l2, _headID) && MatchMotherID(_l1, PDG::ID::JPs) && MatchMotherID(_l2, PDG::ID::JPs);
            if (_option.Contains("PRH")) _cut = _cut && _lep;
            if (_option.Contains("PRL")) _cut = _cut && !_lep;

            if (_sample.BeginsWith("Bs")) _cut = _cut && !TruthMatching(_prj, _ana, _q2bin, "Bs2PhiJPs" + to_string(_ana), _option, _debug, true) && !TruthMatching(_prj, _ana, _q2bin, "Bs2KKJPs" + to_string(_ana), _option, _debug, true);
        }
    }

    if (_debug) MessageSvc::Debug("TruthMatching", &_cut);

    if (_cut == TCut(NOCUT)) MessageSvc::Error("TruthMatching", to_string(_prj), to_string(_ana), _sample, "not supported", "EXIT_FAILURE");

    return _cut;
}

TCut TruthMatching(const ConfigHolder & _configHolder, TString _option, bool _debug) {
    MessageSvc::Info("TruthMatching sample ", _configHolder.GetSample());
    MessageSvc::Info("TruthMatching option ", _option);

    TCut _cut(NOCUT);
    if (_option.Contains("tm") && !_option.Contains("tmSig") && !_option.Contains("tmSwap") && !_option.Contains("tmBkg") && !_option.Contains("tmCustom")) {
        if (_configHolder.IsSignalMC()) {
            MessageSvc::Info("TruthMatching IsSignalMC TRUE ");
            _cut = TRUTHMATCHING::bkgCatSig2;
            if (_option.Contains("tmTight")) _cut = TRUTHMATCHING::bkgCatSignal;
            if (_option.Contains("tmNoGhost")) _cut = TRUTHMATCHING::bkgCatSig1;
        } else {            
            MessageSvc::Info("TruthMatching IsSignalMC FALSE ");
            _cut = TRUTHMATCHING::bkgCatBkg;
        }
        if (_option.Contains("tmIncludeSwap")) _cut = _cut || TRUTHMATCHING::bkgCatSwap;
    } else {
        MessageSvc::Info("CutOption doesnt contain tmStuff ");
        if (_option.Contains("tmSig")) _cut = TRUTHMATCHING::bkgCatSig2;
        if (_option.Contains("tmSigTight")) _cut = TRUTHMATCHING::bkgCatSignal;
        if (_option.Contains("tmSigNoGhost")) _cut = TRUTHMATCHING::bkgCatSig1;
        if (_option.Contains("tmSwap")) _cut = TRUTHMATCHING::bkgCatSwap;
        if (_option.Contains("tmBkg")) _cut = TRUTHMATCHING::bkgCatBkg;
        if (_option.Contains("tmIncludeSwap")) _cut = _cut || TRUTHMATCHING::bkgCatSwap;
        if (_option.Contains("tmCustom")){
    	    //TODO : swap and non swap included?
            if (_option.Contains("tmCustomSwap"))
                _cut = TruthMatching(_configHolder.GetProject(), _configHolder.GetAna(), _configHolder.GetQ2bin(), _configHolder.GetSample(), _option, _debug, true);
            else
                _cut = TruthMatching(_configHolder.GetProject(), _configHolder.GetAna(), _configHolder.GetQ2bin(), _configHolder.GetSample(), _option, _debug, false);	    
        }
    }
    if( _option.Contains("OneGhost")){
        map< pair<Prj, Analysis>, TCut > _extraTMCut = { { {Prj::RK, Analysis::EE},   TCut("( ( (E1_TRUEID==0) + (E2_TRUEID==0) + (K_TRUEID==0) <=1 ) )")},
                                                         { {Prj::RK, Analysis::MM},   TCut("( ( (M1_TRUEID==0) + (M2_TRUEID==0) + (K_TRUEID==0) <=1 ) )")},
                                                         { {Prj::RKst, Analysis::EE}, TCut("( ( (E1_TRUEID==0) + (E2_TRUEID==0) + (K_TRUEID==0) + (Pi_TRUEID==0) <=1 ) )")},
                                                         { {Prj::RKst, Analysis::MM}, TCut("( ( (M1_TRUEID==0) + (M2_TRUEID==0) + (K_TRUEID==0) + (Pi_TRUEID==0) <=1 ) )")},
                                                         { {Prj::RPhi, Analysis::EE}, TCut("( ( (E1_TRUEID==0) + (E2_TRUEID==0) + (K1_TRUEID==0) + (K2_TRUEID==0) <=1 ) )")},
                                                         { {Prj::RPhi, Analysis::MM}, TCut("( ( (M1_TRUEID==0) + (M2_TRUEID==0) + (K1_TRUEID==0) + (K2_TRUEID==0) <=1 ) )")}
        };
        _extraTMCut.at( make_pair( _configHolder.GetProject(), _configHolder.GetAna()));
        _cut = _cut &&  _extraTMCut.at( make_pair( _configHolder.GetProject(), _configHolder.GetAna()));
    }
    if( _option.Contains("OneGhostNMatches") && _option.Contains("tmSig")){    
        TCut _signalLikeness = TCut(Form("SignalLikeness>=%i", _configHolder.GetNBodies()-1));   
    	if(_option.Contains("CustomSignalLikeness")){
            map<  TString, TString > idMatch{   { "L1" , Form("TMath::Abs({L1}_TRUEID) == %i",  _configHolder.GetAna() == Analysis::EE ?  PDG::ID::E : PDG::ID::M) },
                                                { "L2" , Form("TMath::Abs({L2}_TRUEID) == %i",  _configHolder.GetAna() == Analysis::EE ?  PDG::ID::E : PDG::ID::M) }, 
                                                { "H1",  Form("TMath::Abs({H1}_TRUEID) == %i",  PDG::ID::K) } , 
                                                { "H2",  Form("TMath::Abs({H2}_TRUEID) == %i", _configHolder.GetProject() == Prj::RPhi ? PDG::ID::K : PDG::ID::Pi) }
            };
            if( _configHolder.GetAna() == Analysis::EE){
                idMatch["L1"] = idMatch["L1"].ReplaceAll("{L1}","E1");
                idMatch["L2"] = idMatch["L2"].ReplaceAll("{L2}","E2");
            }else if ( _configHolder.GetAna() == Analysis::MM){
                idMatch["L1"] = idMatch["L1"].ReplaceAll("{L1}","M1");
                idMatch["L2"] = idMatch["L2"].ReplaceAll("{L2}","M2");
            }
            if( _configHolder.GetAna() == Analysis::EE) idMatch["L1"] = idMatch["L1"].ReplaceAll("{L1}","E1");
            if( _configHolder.GetProject() == Prj::RK)        _signalLikeness = TCut( Form( " ((%s) + (%s) + (%s)       )  >= 2", idMatch.at("L1").Data(), idMatch.at("L2").Data(), idMatch.at("H1").Data()));	    
            else if( _configHolder.GetProject() == Prj::RKst) _signalLikeness = TCut( Form( " ((%s) + (%s) + (%s) + (%s))  >= 3", idMatch.at("L1").Data(), idMatch.at("L2").Data(), idMatch.at("H1").Data(), idMatch.at("H2").Data()));	    
            else if( _configHolder.GetProject() == Prj::RPhi) _signalLikeness = TCut( Form( " ((%s) + (%s) + (%s) + (%s))  >= 3", idMatch.at("L1").Data(), idMatch.at("L2").Data(), idMatch.at("H1").Data(), idMatch.at("H2").Data()));	  
            else MessageSvc::Warning("Custom Signal Likeness Not Implemented for Project asked");
        }
        auto _add_cut = _signalLikeness;
        _cut = _cut && _add_cut;
    }
    return _cut;
}

#endif
